import datetime
from .rule import Rule
from ..businesstimedelta import localize_unlocalized_dt


class WorkDayRule(Rule):
    """Basic implementation of a working day that starts and ends some days of
    the week at the same start and end time. Ex. Monday through Friday in the EST timezone."""

    def __init__(self, start_time=datetime.time(9), end_time=datetime.time(18),
                 working_days=[0, 1, 2, 3, 4], *args, **kwargs):
        """
        Args:
            start_time: a Time object that defines the start of a work day
            end_time: a Time object that defines the end of a work day
            working_days: days of the working week (0 = Monday)
            tz: a pytz timezone
        """
        kwargs['time_off'] = kwargs.get('time_off', False)
        super(WorkDayRule, self).__init__(*args, **kwargs)
        self.start_time = start_time
        self.end_time = end_time
        self.working_days = working_days

    def next(self, dt, reverse=False):
        dt = localize_unlocalized_dt(dt)
        localized_dt = dt.astimezone(self.tz)

        # Figure out what is the first upcoming working date
        working_date = localized_dt.date()
        if working_date.weekday() in self.working_days \
           and localized_dt.time() < self.end_time:
            pass  # Today is the working date
        else:
            while True:
                working_date += datetime.timedelta(days=1)
                if working_date.weekday() in self.working_days:
                    break

        # We know the target working date now. Just figure out the start and end times.
        start = self.tz.localize(datetime.datetime.combine(working_date, self.start_time))
        end = self.tz.localize(datetime.datetime.combine(working_date, self.end_time))

        # If we are in the range now, set the start or end date to now.
        if start < dt and end > dt:
            start = dt

        return (start, end)

    def previous(self, dt, *args, **kwargs):
        dt = localize_unlocalized_dt(dt)
        localized_dt = dt.astimezone(self.tz)

        # Figure out what is the first upcoming working date
        working_date = localized_dt.date()
        if working_date.weekday() in self.working_days \
           and localized_dt.time() > self.start_time:
            pass  # Today is the working date
        else:
            while True:
                working_date -= datetime.timedelta(days=1)
                if working_date.weekday() in self.working_days:
                    break

        # We know the target working date now. Just figure out the start and end times.
        start = self.tz.localize(datetime.datetime.combine(working_date, self.start_time))
        end = self.tz.localize(datetime.datetime.combine(working_date, self.end_time))

        # If we are in the range now, set the start or end date to now.
        if start < dt and end > dt:
            end = dt

        return (start, end)


class LunchTimeRule(WorkDayRule):
    """Convenience function for lunch breaks."""
    def __init__(self, start_time=datetime.time(12), end_time=datetime.time(13),
                 working_days=[0, 1, 2, 3, 4], *args, **kwargs):
        super(LunchTimeRule, self).__init__(
            start_time=start_time,
            end_time=end_time,
            working_days=working_days,
            *args, **kwargs)
        self.time_off = True



import pytz
import datetime
from ..businesstimedelta import BusinessTimeDelta, localize_unlocalized_dt


class Rule(object):
    """This object defines 'blocks' of time. It can define either working hours
    or an exclusion of working hours (such as holidays, lunch breaks, etc)"""
    def __init__(self, tz=pytz.utc, time_off=False):
        self.tz = tz
        self.time_off = time_off

    def next(self, dt):
        """Returns the start and end of the upcoming (or current) block of time
        that falls within this BusinessTime.

        Args:
            dt: a datetime object.
        Output:
            tuple of (start, end) of the first upcoming business time
            in aware datetime objects.
        """
        raise NotImplementedError

    def previous(self, *args, **kwargs):
        """Same as next, but backwards in time"""
        raise NotImplementedError

    def difference(self, dt1, dt2):
        """Calculate the business time between two datetime objects."""
        dt1 = localize_unlocalized_dt(dt1)
        dt2 = localize_unlocalized_dt(dt2)
        start_dt, end_dt = sorted([dt1, dt2])
        td_sum = datetime.timedelta()
        dt = start_dt

        while True:
            period_start, period_end = self.next(dt)
            period_delta = period_end - period_start

            # If we are past the end_dt, we are done!
            if period_end > end_dt:
                last_day_add = max(end_dt - period_start, datetime.timedelta())
                result = td_sum + last_day_add
                return BusinessTimeDelta(self, hours=result.days * 24, seconds=result.seconds)

            dt = period_end
            td_sum += period_delta



from setuptools import setup

setup(
    name='businesstimedelta',
    version='0.2',
    description="Timedelta for business time. Supports exact amounts of time " +
                "(hours, seconds), custom schedules, holidays, and time zones.",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Office/Business :: Scheduling'
    ],
    keywords='business working time timedelta hours businesstime businesshours',
    url='http://github.com/seppemans/businesstimedelta',
    author='seppemans',
    license='MIT',
    packages=['businesstimedelta', 'businesstimedelta.rules'],
    install_requires=[
        'pytz',
        'holidays'
    ],
    zip_safe=False,
    test_suite='nose.collector',
    tests_require=['nose'])



import datetime
from .rule import Rule
from ..businesstimedelta import localize_unlocalized_dt


class HolidayRule(Rule):
    def __init__(self, holidays, *args, **kwargs):
        """This rule represents a set of holidays.
        Args:
            holidays: a list with dates, or an object from the Holidays python module.
        """
        kwargs['time_off'] = kwargs.get('time_off', True)
        self.holidays = holidays
        super(HolidayRule, self).__init__(*args, **kwargs)

    def __repr__(self):
        return '<HolidayRule: %s>' % (self.holidays)

    def next_holiday(self, date, reverse=False, max_days=365 * 5):
        """ Get the next holiday
        Args:
            date: Find the next holiday after (or at) this date.
            max_days: Allowing both a list of dates as well as an object defined by
                the Holidays module requires a loop to test the holidays object against
                individual dates. To avoid getting stuck in an infinite loop here we need
                to give an upper limit of days to look into the future."""

        count = 0
        while True:
            if date in self.holidays:
                return date

            if reverse:
                date -= datetime.timedelta(days=1)
            else:
                date += datetime.timedelta(days=1)

            count += 1
            if count > max_days:
                return None

    def next(self, dt, reverse=False):
        """Get the start and end of the next holiday after a datetime
        Args:
            dt: datetime
        """
        dt = localize_unlocalized_dt(dt)
        localized_dt = dt.astimezone(self.tz)
        next_holiday = self.next_holiday(localized_dt.date(), reverse=reverse)
        start = self.tz.localize(
            datetime.datetime.combine(
                next_holiday, datetime.time(0, 0, 0)))
        end = start + datetime.timedelta(days=1)

        # If we are in the range now, set the start or end date to now.
        if start < dt and end > dt:
            if reverse:
                end = dt
            else:
                start = dt

        return (start, end)

    def previous(self, *args, **kwargs):
        """Reverse of next function
        """
        kwargs['reverse'] = True
        return self.next(*args, **kwargs)



import datetime
import pytz


def localize_unlocalized_dt(dt):
    """Turn naive datetime objects into UTC.
    Don't do anything if the datetime object is aware.
    https://docs.python.org/3/library/datetime.html#datetime.timezone
    """
    if dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None:
        return dt
    return pytz.utc.localize(dt)


class BusinessTimeDelta(object):
    def __init__(self, rule, hours=0, seconds=0, timedelta=None):
        self.rule = rule

        if timedelta:
            self.timedelta = timedelta
        else:
            self.timedelta = datetime.timedelta(
                seconds=seconds,
                hours=hours)

    def __repr__(self):
        return '<BusinessTimeDelta %s hours %s seconds>' % (self.hours, self.seconds)

    def __eq__(self, other):
        return self.timedelta == other.timedelta

    def __add__(self, other):
        if isinstance(other, BusinessTimeDelta) and other.rule == self.rule:
            return BusinessTimeDelta(self.rule, timedelta=self.timedelta + other.timedelta)

        elif isinstance(other, datetime.datetime):
            dt = localize_unlocalized_dt(other)
            td_left = self.timedelta
            while True:
                period_start, period_end = self.rule.next(dt)
                period_delta = period_end - period_start

                # If we ran out of timedelta, return
                if period_delta >= td_left:
                    return period_start + td_left

                td_left -= period_delta
                dt = period_end

        raise NotImplementedError

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, BusinessTimeDelta) and other.rule == self.rule:
            return BusinessTimeDelta(self.rule, timedelta=self.timedelta - other.timedelta)

        elif isinstance(other, datetime.datetime):
            dt = localize_unlocalized_dt(other)
            td_left = self.timedelta
            while True:
                period_start, period_end = self.rule.previous(dt)
                period_delta = period_end - period_start

                # If we ran out of timedelta, return
                if period_delta >= td_left:
                    return period_end - td_left

                td_left -= period_delta
                dt = period_start

    def __rsub__(self, other):
        return self.__sub__(other)

    @property
    def hours(self):
        return int(self.timedelta.total_seconds() // (60 * 60))

    @property
    def seconds(self):
        return int(self.timedelta.total_seconds() % (60 * 60))



from .rule import Rule
from ..businesstimedelta import localize_unlocalized_dt


class Rules(Rule):
    """Combine a list of rules together to form one rule.
    Args:
        rules: a list of rule objects.
    """
    def __init__(self, rules, *args, **kwargs):
        self.available_rules = [x for x in rules if not x.time_off]
        self.unavailable_rules = [x for x in rules if x.time_off]
        super(Rules, self).__init__(*args, **kwargs)

    def next(self, dt):
        dt = localize_unlocalized_dt(dt)
        min_start = None
        min_end = None

        while True:
            # Find the first upcoming available time
            for rule in self.available_rules:
                start, end = rule.next(dt)

                if not min_start or start < min_start:
                    min_start = start
                    min_end = end

            # Check whether that time is not unavailable due to an
            # unavailability rule. If so, restart this process beginning
            # at the end of this unavailability period.
            for rule in self.unavailable_rules:
                start, end = rule.next(min_start)

                if start == min_start:
                    dt = end
                    min_start = None
                    break

            # We found the first time that is available.
            # Now see when it becomes unavailable.
            if min_start:
                for rule in self.unavailable_rules:
                    start, end = rule.next(min_start)

                    if end < min_end:
                        min_end = start

                if min_end != min_start:
                    return (min_start, min_end)

    def previous(self, dt):
        dt = localize_unlocalized_dt(dt)
        min_start = None
        min_end = None

        while True:
            # Find the first available time in the past
            for rule in self.available_rules:
                start, end = rule.previous(dt)

                if not min_end or end > min_end:
                    min_start = start
                    min_end = end

            # Check whether that time is not unavailable due to an
            # unavailability rule. If so, restart this process beginning
            # at the start of this unavailability period.
            for rule in self.unavailable_rules:
                start, end = rule.previous(min_end)

                if end == min_end:
                    dt = start
                    min_end = None
                    break

            # We found the first time that is available.
            # Now see when it becomes unavailable.
            if min_end:
                for rule in self.unavailable_rules:
                    start, end = rule.previous(min_end)
                    if end > min_start:
                        min_start = end

                if min_end != min_start:
                    return (min_start, min_end)
